body { margin: 0; font: 0.85em "Lucida Grande", "Trebuchet MS", Verdana, sans-serif; color: #fff; background: url(/images/global_bg.gif); }

#page { width: 900px; margin: 0 auto; background: #440008; border-top-width: 5px; border-top-style: solid; border-top-color: #ff8500; }

#header { height: 75px; padding: 0; }
#header h1 { float: left; width: 274px; height: 75px; margin: 0; background-image: url(/images/global_logo.gif); /* Crazy nested comment */ background-repeat: no-repeat; text-indent: -9999px; }
#header .status { float: right; padding-top: 0.5em; padding-left: 0.5em; padding-right: 0.5em; padding-bottom: 0; }
#header .status p { float: left; margin-top: 0; margin-right: 0.5em; margin-bottom: 0; margin-left: 0; }
#header .status ul { float: left; margin: 0; padding: 0; }
#header .status li { list-style-type: none; display: inline; margin: 0 5px; }
#header .status a:link, #header .status a:visited { color: #ff8500; text-decoration: none; }
#header .status a:hover { text-decoration: underline; }
#header .search { float: right; clear: right; margin: 12px 0 0 0; }
#header .search form { margin: 0; }
#header .search input { margin: 0 3px 0 0; padding: 2px; border: none; }

#menu { clear: both; text-align: right; height: 20px; border-bottom: 5px solid #006b95; background: #00a4e4; }
#menu .contests ul { margin: 0 5px 0 0; padding: 0; }
#menu .contests ul li { list-style-type: none; margin: 0 5px; padding: 5px 5px 0 5px; display: inline; font-size: 1.1em; color: #fff; background: #00a4e4; }
#menu .contests a:link, #menu .contests a:visited { color: #fff; text-decoration: none; font-weight: bold; }
#menu .contests a:hover { text-decoration: underline; }

#content { clear: both; }
#content .container { clear: both; }
#content .container .column { float: left; }
#content .container .column .right { float: right; }
#content a:link, #content a:visited { color: #93d700; text-decoration: none; }
#content a:hover { text-decoration: underline; }

#content p, #content div { width: 40em; }
#content p li, #content p dt, #content p dd, #content div li, #content div dt, #content div dd { color: #ddffdd; background-color: #4792bb; }
#content .container.video .column.left { width: 200px; }
#content .container.video .column.left .box { margin-top: 10px; }
#content .container.video .column.left .box p { margin: 0 1em auto 1em; }
#content .container.video .column.left .box.participants img { float: left; margin: 0 1em auto 1em; border: 1px solid #6e000d; border-style: solid; }
#content .container.video .column.left .box.participants h2 { margin: 0 0 10px 0; padding: 0.5em; /* The background image is a gif! */ background: #6e000d url(/images/hdr_participant.gif) 2px 2px no-repeat; /* Okay check this out Multiline comments Wow dude I mean seriously, WOW */ text-indent: -9999px; border-top-width: 5px; border-top-style: solid; border-top-color: #a20013; border-right-width: 1px; border-right-style: dotted; }
#content .container.video .column.middle { width: 500px; }
#content .container.video .column.right { width: 200px; }
#content .container.video .column.right .box { margin-top: 0; }
#content .container.video .column.right .box p { margin: 0 1em auto 1em; }
#content .container.video .column p { margin-top: 0; }

#content.contests .container.information .column.right .box { margin: 1em 0; }
#content.contests .container.information .column.right .box.videos .thumbnail img { width: 200px; height: 150px; margin-bottom: 5px; }
#content.contests .container.information .column.right .box.videos a:link, #content.contests .container.information .column.right .box.videos a:visited { color: #93d700; text-decoration: none; }
#content.contests .container.information .column.right .box.videos a:hover { text-decoration: underline; }
#content.contests .container.information .column.right .box.votes a { display: block; width: 200px; height: 60px; margin: 15px 0; background: url(/images/btn_votenow.gif) no-repeat; text-indent: -9999px; outline: none; border: none; }
#content.contests .container.information .column.right .box.votes h2 { margin: 52px 0 10px 0; padding: 0.5em; background: #6e000d url(/images/hdr_videostats.gif) 2px 2px no-repeat; text-indent: -9999px; border-top: 5px solid #a20013; }

#content.contests .container.video .box.videos h2 { margin: 0; padding: 0.5em; background: #6e000d url(/images/hdr_newestclips.gif) 2px 2px no-repeat; text-indent: -9999px; border-top: 5px solid #a20013; }
#content.contests .container.video .box.videos table { width: 100; }
#content.contests .container.video .box.videos table td { padding: 1em; width: 25; vertical-align: top; }
#content.contests .container.video .box.videos table td p { margin: 0 0 5px 0; }
#content.contests .container.video .box.videos table td a:link, #content.contests .container.video .box.videos table td a:visited { color: #93d700; text-decoration: none; }
#content.contests .container.video .box.videos table td a:hover { text-decoration: underline; }
#content.contests .container.video .box.videos .thumbnail { float: left; }
#content.contests .container.video .box.videos .thumbnail img { width: 80px; height: 60px; margin: 0 10px 0 0; border: 1px solid #6e000d; }

#content .container.comments .column { margin-top: 15px; }
#content .container.comments .column.left { width: 600px; }
#content .container.comments .column.left .box ol { margin: 0; padding: 0; }
#content .container.comments .column.left .box li { list-style-type: none; padding: 10px; margin: 0 0 1em 0; background: #6e000d; border-top: 5px solid #a20013; }
#content .container.comments .column.left .box li div { margin-bottom: 1em; }
#content .container.comments .column.left .box li ul { text-align: right; }
#content .container.comments .column.left .box li ul li { display: inline; border: none; padding: 0; }
#content .container.comments .column.right { width: 290px; padding-left: 10px; }
#content .container.comments .column.right h2 { margin: 0; padding: 0.5em; background: #6e000d url(/images/hdr_addcomment.gif) 2px 2px no-repeat; text-indent: -9999px; border-top: 5px solid #a20013; }
#content .container.comments .column.right .box textarea { width: 290px; height: 100px; border: none; }

#footer { margin-top: 10px; padding: 1.2em 1.5em; background: #ff8500; }
#footer ul { margin: 0; padding: 0; list-style-type: none; }
#footer ul li { display: inline; margin: 0 0.5em; color: #440008; }
#footer ul.links { float: left; }
#footer ul.links a:link, #footer ul.links a:visited { color: #440008; text-decoration: none; }
#footer ul.links a:hover { text-decoration: underline; }
#footer ul.copyright { float: right; }

.clear { clear: both; }

.centered { text-align: center; }

img { border: none; }

button.short { width: 60px; height: 22px; padding: 0 0 2px 0; color: #fff; border: none; background: url(/images/btn_short.gif) no-repeat; }

table { border-collapse: collapse; }
