@import "filename_fn_import";

@mixin local-mixin {
  local-mixin: filename();
}

@function local-function() {
  @return filename();
}

filename {
  local: filename();
  @include local-mixin;
  local-function: local-function();

  @include imported-mixin;
  imported-function: imported-function();
}
