#main,
#header {
  height: 50px; }
  #main div,
  #header div {
    width: 100px; }
    #main div a span,
    #main div em span,
    #header div a span,
    #header div em span {
      color: pink; }

#one div.nested,
#one span.nested,
#one p.nested,
#two div.nested,
#two span.nested,
#two p.nested,
#three div.nested,
#three span.nested,
#three p.nested {
  font-weight: bold;
  border-color: red;
  display: block; }
