/* line 1, ../templates/line_numbers.sass */
foo {
  bar: baz; }

/* line 6, ../templates/importee.sass */
imported {
  otherconst: 12;
  myconst: goodbye; }
  /* line 5, ../templates/line_numbers.sass */
  imported squggle {
    blat: bang; }

/* line 3, ../templates/basic.sass */
body {
  font: Arial;
  background: blue; }

/* line 7, ../templates/basic.sass */
#page {
  width: 700px;
  height: 100; }
  /* line 10, ../templates/basic.sass */
  #page #header {
    height: 300px; }
    /* line 12, ../templates/basic.sass */
    #page #header h1 {
      font-size: 50px;
      color: blue; }

/* line 18, ../templates/basic.sass */
#content.user.show #container.top #column.left {
  width: 100px; }
/* line 20, ../templates/basic.sass */
#content.user.show #container.top #column.right {
  width: 600px; }
/* line 22, ../templates/basic.sass */
#content.user.show #container.bottom {
  background: brown; }

/* line 13, ../templates/importee.sass */
midrule {
  inthe: middle; }

/* line 12, ../templates/line_numbers.sass */
umph {
  foo: bar; }
  /* line 18, ../templates/importee.sass */
  umph baz {
    blat: bang; }
