#main
  :width 15em
  :color #0000ff
  p
    :border
      :style dotted
      /* Nested comment
        More nested stuff
      :width 2px
  .cool
    :width 100px

#left
  :font
    :size 2em
    :weight bold
  :float left
