body
  :margin 0
  :font   0.85em "Lucida Grande", "Trebuchet MS", Verdana, sans-serif
  :color #fff
  :background url(/images/global_bg.gif)

#page
  :width      900px
  :margin     0 auto
  :background #440008
  :border-top
    :width 5px
    :style solid
    :color #ff8500

#header
  :height  75px
  :padding 0
  h1
    :float  left
    :width  274px
    :height 75px
    :margin 0
    :background
      :image url(/images/global_logo.gif)
      /* Crazy nested comment
      :repeat no-repeat
    :text-indent -9999px
  .status
    :float   right
    :padding
      :top .5em
      :left .5em
      :right .5em
      :bottom 0
    p
      :float  left
      :margin
        :top 0
        :right 0.5em
        :bottom 0
        :left 0
    ul
      :float   left
      :margin  0
      :padding 0
    li
      :list-style-type none
      :display inline
      :margin 0 5px
    a:link, a:visited
      :color #ff8500
      :text-decoration none
    a:hover
      :text-decoration underline
  .search
    :float  right
    :clear  right
    :margin 12px 0 0 0
    form
      :margin 0
    input
      :margin 0 3px 0 0
      :padding 2px
      :border none

#menu 
  :clear both
  :text-align right
  :height 20px
  :border-bottom 5px solid #006b95
  :background #00a4e4
  .contests
    ul
      :margin 0 5px 0 0
      :padding 0
      li 
        :list-style-type none
        :margin 0 5px
        :padding 5px 5px 0 5px
        :display inline
        :font-size 1.1em
        // This comment is properly indented
        :color #fff
        :background #00a4e4
    a:link, a:visited
      :color #fff
      :text-decoration none
      :font-weight bold
    a:hover
      :text-decoration underline

//General content information
#content
  :clear both
  .container
    :clear both
    .column
      :float left
      .right
        :float right
  a:link, a:visited
    :color #93d700
    :text-decoration none
  a:hover
    :text-decoration underline

// A hard tab:
	

#content
  p, div
    :width 40em
    li, dt, dd
      :color #ddffdd
      :background-color #4792bb
  .container.video
    .column.left
      :width 200px
      .box
        :margin-top 10px
        p 
          :margin 0 1em auto 1em
      .box.participants 
        img
          :float  left
          :margin 0 1em auto 1em
          :border 1px solid #6e000d
            :style solid
        h2
          :margin 0 0 10px 0
          :padding 0.5em
          /* The background image is a gif!
          :background #6e000d url(/images/hdr_participant.gif) 2px 2px no-repeat
          /* Okay check this out
            Multiline comments
            Wow dude
            I mean seriously, WOW
          :text-indent -9999px
          // And also...
            Multiline comments that don't output!
            Snazzy, no?
          :border
            :top
              :width 5px
              :style solid
              :color #a20013
            :right
              :width 1px
              :style dotted
    .column.middle
      :width 500px
    .column.right
      :width 200px
      .box
        :margin-top 0
        p
          :margin 0 1em auto 1em
    .column
      p
        :margin-top 0

#content.contests
  .container.information 
    .column.right 
      .box
        :margin 1em 0
      .box.videos
        .thumbnail img 
          :width         200px
          :height        150px
          :margin-bottom 5px
        a:link, a:visited
          :color #93d700
          :text-decoration none
        a:hover
          :text-decoration underline
      .box.votes 
        a
          :display block
          :width  200px
          :height 60px
          :margin 15px 0
          :background url(/images/btn_votenow.gif) no-repeat
          :text-indent -9999px
          :outline none
          :border  none
        h2
          :margin  52px 0 10px 0
          :padding 0.5em
          :background #6e000d url(/images/hdr_videostats.gif) 2px 2px no-repeat
          :text-indent -9999px
          :border-top 5px solid #a20013

#content.contests 
  .container.video
    .box.videos
      h2
        :margin  0
        :padding 0.5em
        :background #6e000d url(/images/hdr_newestclips.gif) 2px 2px no-repeat
        :text-indent -9999px
        :border-top 5px solid #a20013
      table
        :width 100
        td 
          :padding 1em
          :width 25
          :vertical-align top
          p
            :margin 0 0 5px 0
          a:link, a:visited    
            :color #93d700
            :text-decoration none
          a:hover
            :text-decoration underline
      .thumbnail
        :float left
        img
          :width  80px
          :height 60px
          :margin 0 10px 0 0
          :border 1px solid #6e000d

#content 
  .container.comments 
    .column
      :margin-top 15px
    .column.left
      :width 600px
      .box 
        ol
          :margin  0
          :padding 0
        li
          :list-style-type none
          :padding 10px
          :margin 0 0 1em 0
          :background #6e000d
          :border-top 5px solid #a20013
          div
            :margin-bottom 1em
          ul
            :text-align right
            li
              :display inline
              :border none
              :padding 0
    .column.right
      :width        290px
      :padding-left 10px
      h2
        :margin 0
        :padding 0.5em
        :background #6e000d url(/images/hdr_addcomment.gif) 2px 2px no-repeat
        :text-indent -9999px
        :border-top 5px solid #a20013
      .box
        textarea
          :width  290px
          :height 100px
          :border none

#footer
  :margin-top 10px
  :padding    1.2em 1.5em
  :background #ff8500
  ul
    :margin 0
    :padding 0
    :list-style-type none
    li
      :display inline
      :margin  0 0.5em
      :color   #440008
  ul.links
    :float left
    a:link, a:visited
      :color #440008
      :text-decoration none
    a:hover
      :text-decoration underline
  ul.copyright
    :float right


.clear 
  :clear both

.centered 
  :text-align center

img
  :border none

button.short 
  :width   60px
  :height  22px
  :padding 0 0 2px 0
  :color   #fff
  :border  none
  :background url(/images/btn_short.gif) no-repeat

table
  :border-collapse collapse
