h1
  :float  left
  :width  274px
  height: 75px
  margin: 0
  background:
    repeat: no-repeat
    :image none
  a:hover, a:visited
    color: green
  b:hover
    color: red
    :background-color green
  const
    nosp: 1 + 2
    sp : 1 + 2
