$postconst: goodbye

=postmixin
  post-mixin: here

imported
  :otherconst $preconst
  :myconst $postconst
  +premixin

@import basic

midrule
  :inthe middle

=crazymixin
  foo: bar
  baz
    blat: bang
