@charset "IBM866"

.foo
  a: b

@import "foo.css"

// Even though the imported file is in UTF-8,
// since the root file is in IBM866
// the output will end up being IBM866.
@import "imported_charset_utf8"
