.maruku_toc
  background: #ddd
  border: 1px solid #ccc
  margin-right: 2em
  float: left
  ul
    padding: 0 1em
  #frequently_asked_questions + &
    float: none
    margin: 0 2em

#filecontents
  *:target, dt:target + dd
    background-color: #ccf
    border: 1px solid #88b
  dt
    font-weight: bold
  dd
    margin:
      left: 0
      bottom: 0.7em
    padding-left: 3em
  dt:target
    border-bottom-style: none
    & + dd
      border-top-style: none
