#main,
#header
  height: 50px
  div
    width: 100px
    a,
    em
      span
        color: pink

#one,
#two,
#three
  div.nested,
  span.nested,
  p.nested
    :font
      :weight bold
    :border-color red
    :display block