a { color: #000; }
a:hover { color: #f00; }

p, div { width: 100em; }
p foo, div foo { width: 10em; }
p:hover, p bar, div:hover, div bar { height: 20em; }

#cool { border-style: solid; border-width: 2em; }
.ie7 #cool, .ie6 #cool { content: string("Totally not cool."); }
.firefox #cool { content: string("Quite cool."); }

.wow, .snazzy { font-family: fantasy; }
.wow:hover, .wow:visited, .snazzy:hover, .snazzy:visited { font-weight: bold; }
