#main {
  width: 15em;
  color: #0000ff; }
  #main p {
    border-style: dotted;
    /* Nested comment
     * More nested stuff */
    border-width: 2px; }
  #main .cool {
    width: 100px; }

#left {
  font-size: 2em;
  font-weight: bold;
  float: left; }

#right .header {
  border-style: solid; }
#right .body {
  border-style: dotted; }
#right .footer {
  border-style: dashed; }
